@@ -20,6 +20,7 @@
 
 from __future__ import division, print_function, absolute_import
 
+import sys
 import time
 import functools
 import traceback
@@ -28,8 +29,9 @@ from collections import namedtuple
 
 import hypothesis.strategies as st
 from hypothesis.errors import Flaky, Timeout, NoSuchExample, \
-    Unsatisfiable, InvalidArgument, FailedHealthCheck, \
-    UnsatisfiedAssumption, HypothesisDeprecationWarning
+    Unsatisfiable, InvalidArgument, DeadlineExceeded, MultipleFailures, \
+    FailedHealthCheck, UnsatisfiedAssumption, \
+    HypothesisDeprecationWarning
 from hypothesis.control import BuildContext
 from hypothesis._settings import settings as Settings
 from hypothesis._settings import Phase, Verbosity, HealthCheck, \
@@ -38,20 +40,20 @@ from hypothesis.executors import new_style_executor, \
     default_new_style_executor
 from hypothesis.reporting import report, verbose_report, current_verbosity
 from hypothesis.statistics import note_engine_for_statistics
-from hypothesis.internal.compat import str_to_bytes, get_type_hints, \
-    getfullargspec
-from hypothesis.utils.conventions import infer
+from hypothesis.internal.compat import ceil, str_to_bytes, \
+    get_type_hints, getfullargspec
+from hypothesis.utils.conventions import infer, not_set
 from hypothesis.internal.escalation import \
     escalate_hypothesis_internal_error
-from hypothesis.internal.reflection import is_mock, nicerepr, arg_string, \
-    impersonate, function_digest, fully_qualified_name, \
+from hypothesis.internal.reflection import is_mock, proxies, nicerepr, \
+    arg_string, impersonate, function_digest, fully_qualified_name, \
     define_function_signature, convert_positional_arguments, \
     get_pretty_function_description
 from hypothesis.internal.conjecture.data import Status, StopTest, \
     ConjectureData
 from hypothesis.searchstrategy.strategies import SearchStrategy
 from hypothesis.internal.conjecture.engine import ExitReason, \
-    ConjectureRunner, uniform
+    ConjectureRunner, uniform, sort_key
 
 
 def new_random():
@@ -503,13 +505,44 @@ class StateForActualGivenExecution(object):
     def __init__(self, test_runner, search_strategy, test, settings, random):
         self.test_runner = test_runner
         self.search_strategy = search_strategy
-        self.test = test
         self.settings = settings
         self.at_least_one_success = False
         self.last_exception = None
         self.repr_for_last_exception = None
-        self.falsifying_example = None
+        self.falsifying_examples = ()
+        self.__was_flaky = False
         self.random = random
+        self.__warned_deadline = False
+        if self.settings.deadline is None:
+            self.test = test
+        else:
+            @proxies(test)
+            def timed_test(*args, **kwargs):
+                start = time.time()
+                result = test(*args, **kwargs)
+                runtime = (time.time() - start) * 1000
+                if self.settings.deadline is not_set:
+                    if (
+                        not self.__warned_deadline and
+                        runtime >= 200
+                    ):
+                        self.__warned_deadline = True
+                        note_deprecation((
+                            'Test took %.2fms to run. In future the default '
+                            'deadline setting will be 200ms, which will '
+                            'make this an error. You can set deadline to '
+                            'an explicit value of e.g. %d to turn tests '
+                            'slower than this into an error, or you can set '
+                            'it to None to disable this check entirely.') % (
+                                runtime, ceil(runtime / 100) * 100,
+                        ))
+                elif runtime >= self.settings.deadline:
+                    raise DeadlineExceeded((
+                        'Test took %.2fms, which exceeds the deadline of '
+                        '%.2fms') % (runtime, self.settings.deadline)
+                    )
+                return result
+            self.test = timed_test
 
     def evaluate_test_data(self, data):
         if (
@@ -541,9 +574,15 @@ class StateForActualGivenExecution(object):
             raise
         except Exception:
             escalate_hypothesis_internal_error()
-            self.last_exception = traceback.format_exc()
-            verbose_report(self.last_exception)
-            data.mark_interesting()
+            data.__expected_exception = traceback.format_exc()
+            verbose_report(data.__expected_exception)
+
+            error_class, _, tb = sys.exc_info()
+
+            origin = traceback.extract_tb(tb)[-1]
+            filename = origin[0]
+            lineno = origin[1]
+            data.mark_interesting((error_class, filename, lineno))
 
     def run(self):
         # Tell pytest to omit the body of this function from tracebacks
@@ -561,8 +600,11 @@ class StateForActualGivenExecution(object):
         timed_out = runner.exit_reason == ExitReason.timeout
         if runner.last_data is None:
             return
-        if runner.last_data.status == Status.INTERESTING:
-            self.falsifying_example = runner.last_data.buffer
+        if runner.interesting_examples:
+            self.falsifying_examples = sorted(
+                [d for d in runner.interesting_examples.values()],
+                key=lambda d: sort_key(d.buffer), reverse=True
+            )
         else:
             if timed_out:
                 note_deprecation((
@@ -603,50 +645,89 @@ class StateForActualGivenExecution(object):
                         get_pretty_function_description(self.test),
                         runner.valid_examples,))
 
-        if self.falsifying_example is None:
+        if not self.falsifying_examples:
             return
 
-        assert self.last_exception is not None
+        flaky = 0
 
-        try:
-            with self.settings:
-                self.test_runner(
-                    ConjectureData.for_buffer(self.falsifying_example),
-                    reify_and_execute(
-                        self.search_strategy, self.test,
-                        print_example=True, is_final=True
-                    ))
-        except (UnsatisfiedAssumption, StopTest):
-            report(traceback.format_exc())
-            raise Flaky(
-                'Unreliable assumption: An example which satisfied '
-                'assumptions on the first run now fails it.'
-            )
+        for falsifying_example in self.falsifying_examples:
+            self.__was_flaky = False
+            raised_exception = False
+            try:
+                with self.settings:
+                    self.test_runner(
+                        ConjectureData.for_buffer(falsifying_example.buffer),
+                        reify_and_execute(
+                            self.search_strategy, self.test,
+                            print_example=True, is_final=True
+                        ))
+            except (UnsatisfiedAssumption, StopTest):
+                report(traceback.format_exc())
+                self.__flaky(
+                    'Unreliable assumption: An example which satisfied '
+                    'assumptions on the first run now fails it.'
+                )
+            except:
+                if len(self.falsifying_examples) <= 1:
+                    raise
+                raised_exception = True
+                report(traceback.format_exc())
 
-        report(
-            'Failed to reproduce exception. Expected: \n' +
-            self.last_exception,
-        )
+            if not raised_exception:
+                report(
+                    'Failed to reproduce exception. Expected: \n' +
+                    falsifying_example.__expected_exception,
+                )
 
-        filter_message = (
-            'Unreliable test data: Failed to reproduce a failure '
-            'and then when it came to recreating the example in '
-            'order to print the test data with a flaky result '
-            'the example was filtered out (by e.g. a '
-            'call to filter in your strategy) when we didn\'t '
-            'expect it to be.'
-        )
+                filter_message = (
+                    'Unreliable test data: Failed to reproduce a failure '
+                    'and then when it came to recreating the example in '
+                    'order to print the test data with a flaky result '
+                    'the example was filtered out (by e.g. a '
+                    'call to filter in your strategy) when we didn\'t '
+                    'expect it to be.'
+                )
 
-        try:
-            self.test_runner(
-                ConjectureData.for_buffer(self.falsifying_example),
-                reify_and_execute(
-                    self.search_strategy,
-                    test_is_flaky(self.test, self.repr_for_last_exception),
-                    print_example=True, is_final=True
-                ))
-        except (UnsatisfiedAssumption, StopTest):
-            raise Flaky(filter_message)
+                try:
+                    self.test_runner(
+                        ConjectureData.for_buffer(falsifying_example.buffer),
+                        reify_and_execute(
+                            self.search_strategy,
+                            test_is_flaky(
+                                self.test, self.repr_for_last_exception),
+                            print_example=True, is_final=True
+                        ))
+                except (UnsatisfiedAssumption, StopTest):
+                    self.__flaky(filter_message)
+                except Flaky as e:
+                    if len(self.falsifying_examples) > 1:
+                        self.__flaky(e.args[0])
+                    else:
+                        raise
+
+            if self.__was_flaky:
+                flaky += 1
+
+        # If we only have one example then we should have raised an error or
+        # flaky prior to this point.
+        assert len(self.falsifying_examples) > 1
+
+        if flaky > 0:
+            raise Flaky((
+                'Hypothesis found %d distinct failures, but %d of them '
+                'exhibited some sort of flaky behaviour.') % (
+                    len(self.falsifying_examples), flaky))
+        else:
+            raise MultipleFailures((
+                'Hypothesis found %d distinct failures.') % (
+                    len(self.falsifying_examples,)))
+
+    def __flaky(self, message):
+        if len(self.falsifying_examples) <= 1:
+            raise Flaky(message)
+        else:
+            self.__was_flaky = True
+            report('Flaky example! ' + message)
 
 
 def given(*given_arguments, **given_kwargs):
